import asyncio
import json
from dataclasses import dataclass
from typing import Any, Awaitable, Callable, List

from ocpp.messages import MessageType
from ocpp.v16 import call as v16_call
from ocpp.v16 import call_result as v16_call_result
from ocpp.v20 import call as v20_call
from ocpp.v20 import call_result as v20_call_result
from ocpp.v201 import call as v201_call
from ocpp.v201 import call_result as v201_call_result

from ocpp_asgi.asgi import (
    ASGIHTTPEvent,
    ASGILifeSpanEvent,
    ASGILifeSpanShutDown,
    ASGILifeSpanStartup,
    ASGIScope,
    ASGIWebSocketEvent,
    Receive,
    Scope,
    Send,
)
from ocpp_asgi.logging import log
from ocpp_asgi.router import OCPPAdapter, Router, RouterContext, Subprotocol

ocpp_adapters = {
    Subprotocol.ocpp201.value: OCPPAdapter(
        call=v201_call, call_result=v201_call_result, ocpp_version="2.0.1"
    ),
    Subprotocol.ocpp20.value: OCPPAdapter(
        call=v20_call, call_result=v20_call_result, ocpp_version="2.0"
    ),
    Subprotocol.ocpp16.value: OCPPAdapter(
        call=v16_call, call_result=v16_call_result, ocpp_version="1.6"
    ),
}


@dataclass
class HTTPEventContext:
    charging_station_id: str
    subprotocols: List[str]
    body: dict


@dataclass
class SendAdapter:
    send: Awaitable[Any]
    on_receive: Callable[[str, RouterContext], Awaitable[None]]
    http_from_server_to_client: Callable[[str, RouterContext], Awaitable[str]]
    scope: dict

    async def __call__(self, message: str, is_response: bool, context: RouterContext):
        if is_response:
            if self.scope["type"] == ASGIScope.websocket:
                await self.send(
                    {"type": ASGIWebSocketEvent.send.value, "text": message}
                )
            else:
                await self.send(
                    {"type": ASGIHTTPEvent.response_start.value, "status": 200}
                )
                await self.send(
                    {
                        "type": ASGIHTTPEvent.response_body.value,
                        "body": message.encode("utf-8"),
                    }
                )
        else:
            if self.scope["type"] == ASGIScope.websocket:
                await self.send(
                    {"type": ASGIWebSocketEvent.send.value, "text": message}
                )
            else:
                log.debug(f"<- HTTP: {context.charging_station_id=} {message=}")
                await self.http_from_server_to_client(message=message, context=context)


class ASGIApplication:
    """ASGI Application to handle event based message routing."""

    def __init__(self):
        self.routers: Router = {}

    def include_router(self, router: Router):
        self.routers[router.subprotocol] = router

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        """ASGI signature handler.

        Args:
            scope (Scope): ASGI scope
            receive (Receive): ASGI handle for receiving messages
            send (Send): ASGI handle for sending messages
        """

        if scope["type"] in [ASGIScope.websocket, ASGIScope.http]:
            await self.handler(scope, receive, send)
        elif scope["type"] == ASGIScope.lifespan:
            await self.lifespan_handler(scope, receive, send)
        else:
            raise ValueError(f'Unsupported ASGI scope type: {scope["type"]}')

    async def lifespan_handler(
        self, scope: Scope, receive: Receive, send: Send
    ) -> None:
        event = await receive()
        if event["type"] == ASGILifeSpanEvent.startup.value:
            try:
                await self.on_startup()
                await send({"type": ASGILifeSpanStartup.complete.value})
            except Exception:
                await send({"type", ASGILifeSpanStartup.failed.value})
        elif event["type"] == ASGILifeSpanEvent.shutdown.value:
            try:
                await self.on_shutdown()
                await send({"type": ASGILifeSpanShutDown.complete.value})
            except Exception:
                await send({"type", ASGILifeSpanShutDown.failed.value})

    async def handler(self, scope: Scope, receive: Receive, send: Send):
        log.debug(f"{scope=}")
        while True:
            event = await receive()
            log.debug(f"{event=}")
            context: RouterContext = self._create_context(
                scope=scope, event=event, send=send
            )

            # WebSocket
            if event["type"] == ASGIWebSocketEvent.receive:
                await self.on_receive(message=context.body, context=context)
            elif event["type"] == ASGIWebSocketEvent.connect:
                response = await self.on_connect(context)
                if response:
                    await send({"type": "websocket.accept"})
                else:
                    await send({"type": "websocket.reject"})
            elif event["type"] == ASGIWebSocketEvent.disconnect:
                await self.on_disconnect(
                    charging_station_id=context.charging_station_id,
                    subprotocol=context.subprotocol,
                    code=event["code"],
                )
                break

            # HTTP
            elif event["type"] == ASGIHTTPEvent.request:
                if context is None:
                    await send(
                        {"type": ASGIHTTPEvent.response_start.value, "status": 400}
                    )
                    await send({"type": ASGIHTTPEvent.response_body.value})
                    break
                # TODO: handle more_body case
                message_type = int(context.body[1])
                if message_type != MessageType.Call:
                    # For "CallResult" and "CallError" send empty response back
                    # as ocpp protocol doesn't mandate response for these message types
                    # For "Call" response will be sent by router
                    await send(
                        {"type": ASGIHTTPEvent.response_start.value, "status": 200}
                    )
                    await send({"type": ASGIHTTPEvent.response_body.value})
                await self.on_receive(message=context.body, context=context)
            elif event["type"] == ASGIHTTPEvent.disconnect.value:
                break

    async def on_receive(self, *, message: str, context: RouterContext):
        router: Router = self.routers[context.subprotocol]
        try:
            await router.route_message(message=message, context=context)
        except Exception as e:
            log.error(f"Failure when processing message on_receive: {e=}")
            pass

    # Handlers to override in subclass

    async def on_startup(self):
        """ASGI lifecycle event"""
        pass

    async def on_shutdown(self):
        """ASGI lifecycle event"""
        pass

    async def on_connect(self, context: RouterContext) -> bool:
        """Invoked when websocket connection is being established.

        @return bool: True if connection is allowed, False if rejected.
        """
        return True

    async def on_disconnect(
        self, *, charging_station_id: str, subprotocol: Subprotocol, code: int
    ):
        """Invoked when websocket connection is disconnected."""
        pass

    def http_parse_event(self, http_event: dict) -> HTTPEventContext:
        """Parse context and content from http event's body.

        Relevant only for HTTP central system.

        ASGI event will carry the content in event["body"].
        This is application specific adaptation how data is structured in body.
        """
        raise NotImplementedError

    async def http_from_server_to_client(self, message: str, context: RouterContext):
        """Handle sending message from http backend towards WebSocket proxy.

        Relevant only for HTTP central system.

        HTTP backend needs to send Call type messages to WebSocket proxy.
        This is application specific adaptation how this sending is done.
        """
        raise NotImplementedError

    # Private

    def _create_context(
        self, *, scope: Scope, event: dict, send: Send
    ) -> RouterContext or None:
        queue = asyncio.Queue()
        call_lock = asyncio.Lock()
        subprotocols: list[str] = []
        if scope["type"] == ASGIScope.websocket:
            charging_station_id = scope["path"].strip("/")
            subprotocols = scope["subprotocols"]
            body = event["text"] if "text" in event else None
        else:  # scope["type"] == ASGIScope.http:
            if event["type"] == ASGIHTTPEvent.disconnect:
                # Don't bother creating context for disconnect as it's not used
                return None
            if len(event["body"]) > 0:
                http_event = json.loads(event["body"])
                http_event_context: HTTPEventContext = self.http_parse_event(http_event)
                charging_station_id = http_event_context.charging_station_id
                subprotocols = http_event_context.subprotocols
                body = http_event_context.body

        if len(subprotocols) == 0:
            return None

        # Pick the highest matching subprotocol
        if Subprotocol.ocpp201 in subprotocols:
            subprotocol = Subprotocol.ocpp201.value
        elif Subprotocol.ocpp20 in subprotocols:
            subprotocol = Subprotocol.ocpp20.value
        elif Subprotocol.ocpp16 in subprotocols:
            subprotocol = Subprotocol.ocpp16.value
        else:
            raise ValueError

        send_adapter = SendAdapter(
            scope=scope,
            send=send,
            on_receive=self.on_receive,
            http_from_server_to_client=self.http_from_server_to_client,
        )
        context = RouterContext(
            scope=scope,
            # store parsed body to avoid parsing twice for on_action and after_action
            body=body,
            subprotocol=subprotocol,
            ocpp_adapter=ocpp_adapters[subprotocol],
            send=send_adapter,
            charging_station_id=charging_station_id,
            queue=queue,
            call_lock=call_lock,
        )
        return context



from __future__ import annotations

import asyncio
import functools
import inspect
import uuid
from dataclasses import asdict, dataclass
from enum import Enum
from typing import Any, Awaitable, Callable, Optional

from ocpp.charge_point import camel_to_snake_case, remove_nones, snake_to_camel_case
from ocpp.exceptions import NotImplementedError, OCPPError
from ocpp.messages import Call, MessageType, unpack, validate_payload

from ocpp_asgi.logging import log


class Subprotocol(str, Enum):
    ocpp16 = "ocpp1.6"
    ocpp20 = "ocpp2.0"
    ocpp201 = "ocpp2.0.1"


@dataclass
class OCPPAdapter:
    """OCPPAdapter encapsulates OCPP version specific call and call_result methods."""

    ocpp_version: str
    call: Awaitable[Any]
    call_result: Awaitable[Any]


@dataclass
class RouterContext:
    """RouterContext instance is passed to router."""

    scope: dict  # Store ASGI scope dictionary as is
    body: dict  # Store ASGI content mainly to avoid parsing http event twice
    subprotocol: Subprotocol
    ocpp_adapter: OCPPAdapter
    send: Callable[[str, bool, RouterContext], Awaitable[None]]
    charging_station_id: str
    queue: Any
    call_lock: Any


@dataclass
class HandlerContext:
    """HandlerContext instance is passed to handler."""

    charging_station_id: str
    # References to RouterContext and Router added here so that
    # we can send messages to specific Charging Station, which initiated messaging.
    _router_context: RouterContext
    _router: Router

    async def send(self, message: dataclass) -> Any:
        """Send message to Charging Station within action handler."""
        # Use a lock to prevent make sure that only 1 message can be send at a
        # a time.
        async with self._router_context.call_lock:
            return await self._router.call(
                message=message, context=self._router_context
            )


def subprotocol_to_ocpp_version(subprotocol: str) -> str:
    """Strip away ocpp prefix from"""
    return subprotocol[4:]


class Router:
    """Router is a collection of ocpp action handlers."""

    subprotocol: Subprotocol = None

    def __init__(
        self,
        *,
        subprotocol: Subprotocol,
        response_timeout: Optional[int] = 30,
        create_task: bool = True,
    ):
        """Initialize Router instance.

        Args:
            subprotocol (Subprotocol): Defines the ocpp protocol version for this router
            response_timeout (int): When no response on a request is received
                within this interval, a asyncio.TimeoutError is raised.
            create_task (bool): Create asyncio.Task for executing
                "after"-handler. Does not affect "on-handler".
        """
        self.subprotocol = subprotocol

        # The maximum time in seconds it may take for a CP to respond to a
        # CALL. An asyncio.TimeoutError will be raised if this limit has been
        # exceeded.
        self._response_timeout = response_timeout

        # A dictionary that hooks for Actions. So if the CS receives a it will
        # look up the Action into this map and execute the corresponding hooks
        # if exists.
        # Dictionary contains the following structure for each Action:
        # {
        #     Action.BootNotification: {
        #         "_on_action": <reference to "on_boot_notification">,
        #         "_after_action": <reference to "after_boot_notification">,
        #         "_skip_schema_validation": False,
        #     },
        # }
        self._route_map = {}

        # Function used to generate unique ids for CALLs. By default
        # uuid.uuid4() is used, but it can be changed. This is meant primarily
        # for testing purposes to have predictable unique ids.
        self._unique_id_generator = uuid.uuid4

        # Use asyncio.create_task for "after"-handler.
        self._create_task = create_task
        self._response_queue = asyncio.Queue()

        # Dictionary for storing subscribers for, which are waiting for CallResult or
        # CallErrors.
        self.subscriptions = {}

    def on(self, action, *, skip_schema_validation=False):
        def decorator(func):
            @functools.wraps(func)
            def inner(*args, **kwargs):
                return func(*args, **kwargs)

            option = "_on_action"
            if action not in self._route_map:
                self._route_map[action] = {}
            self._route_map[action][option] = func
            self._route_map[action]["_skip_schema_validation"] = skip_schema_validation
            return inner

        return decorator

    def after(self, action):
        def decorator(func):
            @functools.wraps(func)
            def inner(*args, **kwargs):
                return func(*args, **kwargs)

            option = "_after_action"
            if action not in self._route_map:
                self._route_map[action] = {}
            self._route_map[action][option] = func
            return inner

        return decorator

    async def route_message(self, *, message: str, context: RouterContext):
        """
        Route a message received from a Charging Station.

        If the message is a of type Call the corresponding hooks are executed.
        If the message is of type CallResult or CallError the message is passed
        to the call() function via the response_queue.
        """
        try:
            msg = unpack(message)
        except OCPPError as e:
            log.exception(
                "Unable to parse message: '%s', it doesn't seem "
                "to be valid OCPP: %s",
                message,
                e,
            )
            return

        if msg.message_type_id == MessageType.Call:
            await self._handle_call(msg, context=context)

        elif msg.message_type_id in [
            MessageType.CallResult,
            MessageType.CallError,
        ]:
            if msg.unique_id in self.subscriptions:
                self.subscriptions[msg.unique_id].put_nowait(msg)

    async def _handle_call(self, msg, *, context: RouterContext = None):
        """
        Execute all hooks installed for based on the Action of the message.

        First the '_on_action' hook is executed and its response is returned to
        the client. If there is no '_on_action' hook for Action in the message
        a CallError with a NotImplemtendError is returned.

        Next the '_after_action' hook is executed.

        """
        ocpp_version = subprotocol_to_ocpp_version(self.subprotocol)

        try:
            handlers = self._route_map[msg.action]
        except KeyError:
            raise NotImplementedError(f"No handler for '{msg.action}' " "registered.")

        if not handlers.get("_skip_schema_validation", False):
            validate_payload(msg, ocpp_version)

        # OCPP uses camelCase for the keys in the payload. It's more pythonic
        # to use snake_case for keyword arguments. Therefore the keys must be
        # 'translated'. Some examples:
        #
        # * chargePointVendor becomes charge_point_vendor
        # * firmwareVersion becomes firmwareVersion
        snake_case_payload = camel_to_snake_case(msg.payload)

        try:
            handler = handlers["_on_action"]
        except KeyError:
            raise NotImplementedError(f"No handler for '{msg.action}' " "registered.")

        handler_context = HandlerContext(
            charging_station_id=context.charging_station_id,
            _router_context=context,
            _router=self,
        )
        # Convert message to correct Call instance
        class_ = getattr(context.ocpp_adapter.call, f"{msg.action}Payload")
        payload = class_(**snake_case_payload)
        try:
            response = handler(payload=payload, context=handler_context)
            if inspect.isawaitable(response):
                response = await response
        except Exception as e:
            log.exception("Error while handling request '%s'", msg)
            response = msg.create_call_error(e).to_json()
            await self._send(message=response, is_response=True, context=context)

        temp_response_payload = asdict(response)

        # Remove nones ensures that we strip out optional arguments
        # which were not set and have a default value of None
        response_payload = remove_nones(temp_response_payload)

        # The response payload must be 'translated' from snake_case to
        # camelCase. So:
        #
        # * charge_point_vendor becomes chargePointVendor
        # * firmware_version becomes firmwareVersion
        camel_case_payload = snake_to_camel_case(response_payload)

        response = msg.create_call_result(camel_case_payload)

        if not handlers.get("_skip_schema_validation", False):
            validate_payload(response, ocpp_version)

        await self._send(message=response.to_json(), is_response=True, context=context)

        try:
            handler = handlers["_after_action"]
            response = handler(payload=payload, context=handler_context)
            if inspect.isawaitable(response):
                if self._create_task:
                    # Create task to avoid blocking when making a call
                    # inside the after handler
                    asyncio.ensure_future(response)
                else:
                    await response
        except KeyError:
            # '_on_after' hooks are not required. Therefore ignore exception
            # when no '_on_after' hook is installed.
            pass

    async def call(self, *, message: Any, context: RouterContext):
        ocpp_version = subprotocol_to_ocpp_version(self.subprotocol)

        camel_case_payload = snake_to_camel_case(asdict(message))

        call = Call(
            unique_id=str(self._unique_id_generator()),
            action=message.__class__.__name__[:-7],
            payload=remove_nones(camel_case_payload),
        )

        validate_payload(call, ocpp_version)

        await self._send(message=call.to_json(), is_response=False, context=context)
        self.subscriptions[call.unique_id] = context.queue
        try:
            response = await asyncio.wait_for(
                context.queue.get(), self._response_timeout
            )
        except asyncio.TimeoutError:
            del self.subscriptions[call.unique_id]
            raise

        if response.message_type_id == MessageType.CallError:
            log.warning("Received a CALLError: %s'", response)
            raise response.to_exception()
        else:
            response.action = call.action
            validate_payload(response, ocpp_version)

        snake_case_payload = camel_to_snake_case(response.payload)
        call_result = context.ocpp_adapter.call_result
        cls = getattr(call_result, message.__class__.__name__)
        return cls(**snake_case_payload)

    async def _send(self, *, message: str, is_response: bool, context: RouterContext):
        log.debug(f"{context.charging_station_id=} {message=}")
        await context.send(message=message, is_response=is_response, context=context)



from enum import Enum
from typing import Any, Awaitable, Callable, MutableMapping


class ASGIScope(str, Enum):
    lifespan = "lifespan"
    websocket = "websocket"
    http = "http"


class ASGILifeSpanEvent(str, Enum):
    startup = "lifespan.startup"
    shutdown = "lifespan.shutdown"


class ASGIWebSocketEvent(str, Enum):
    # receive
    connect = "websocket.connect"
    receive = "websocket.receive"
    disconnect = "websocket.disconnect"
    # send
    accept = "websocket.accept"
    send = "websocket.send"
    close = "websocket.close"


class ASGIHTTPEvent(str, Enum):
    disconnect = "http.disconnect"
    request = "http.request"
    response_start = "http.response.start"
    response_body = "http.response.body"


class ASGILifeSpanStartup(str, Enum):
    complete = "lifespan.startup.complete"
    failed = "lifespan.startup.failed"


class ASGILifeSpanShutDown(str, Enum):
    complete = "lifespan.shutdown.complete"
    failed = "lifespan.shutdown.failed"


Scope = MutableMapping[str, Any]
Message = MutableMapping[str, Any]

Receive = Callable[[], Awaitable[Message]]
Send = Callable[[Message], Awaitable[None]]

ASGIAppSignature = Callable[[Scope, Receive, Send], Awaitable[None]]



from ocpp.messages import Call, CallError, unpack


def create_call_error(message: str) -> str:
    """Create CallError serialized representation based on serialize Call.

    Raises ValueError if message is not type Call. CallResult and CallError
    don't require response.
    """
    call: Call = unpack(message)
    if isinstance(call, Call):
        call_error: CallError = call.create_call_error(None)
        return call_error.to_json()
    else:
        raise ValueError("message is not type Call")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ocpp_asgi']

package_data = \
{'': ['*']}

install_requires = \
['ocpp>=0.15.0,<0.16.0']

setup_kwargs = {
    'name': 'ocpp-asgi',
    'version': '0.3.0',
    'description': 'ocpp-asgi provides ASGI compliant interface for implementing event-driven server-side support for OCPP protocol with Python',
    'long_description': '<p align="center">\n<a href="https://github.com/villekr/ocpp-asgi/actions?query=workflow%3Amain" target="_blank">\n    <img src="https://github.com/villekr/ocpp-asgi/workflows/main/badge.svg" alt="main">\n</a>\n<a href="https://codecov.io/gh/villekr/ocpp-asgi">\n  <img src="https://codecov.io/gh/villekr/ocpp-asgi/branch/main/graph/badge.svg?token=DZ2QWVF3DX"/>\n</a>\n<a href="https://pypi.org/project/ocpp-asgi" target="_blank">\n    <img src="https://img.shields.io/pypi/v/ocpp-asgi?color=%2334D058&label=pypi%20package" alt="Package version">\n</a>\n</p>\n\n---\n\n# OCPP-ASGI\n\nocpp-asgi provides **ASGI compliant** interface for implementing **event-driven** **server-side** support for OCPP protocol with Python. It depends on and extends [mobilityhouse/ocpp](https://github.com/mobilityhouse/ocpp). \n\nThe key features are:\n* ASGI compliant interface supporting both WebSocket and HTTP protocols.\n* Event-driven and "stateless" approach for implementing action handlers for OCPP messages. \n* Highly-scalable and supports serverless (e.g. AWS Lambda, Azure Functions) with compatible ASGI server.\n* Requires small and straightforward changes from ocpp to action handlers (but is not backwards compatible).\n\nRead the [blog post](https://ville-karkkainen.medium.com/towards-event-based-serverless-ocpp-backend-system-part-i-motivation-and-architecture-options-5d51ba09dfd6) about the motivation behind creating this library.\n\n## Disclaimer!\n\nThis library is still in alpha state. At the moment I don\'t have time nor immediate business need to invest in further development. However, I\'m happy to discuss and help if you are considering the benefits of using this library in you project.\n\nPlease send [email](mailto:ville.karkkainen@outlook.com) if you have any questions about this library or have some business inquiries in mind.\n\n# Getting started\n\n## Installation\n\n```\npip install ocpp-asgi\n```\n\nAlso ASGI server is required e.g. [uvicorn](https://www.uvicorn.org) or [mangum](https://www.uvicorn.org) when deployed to AWS Lambda with API Gateway.\n```\npip install uvicorn\n```\n\n## Action handlers\n\nDecorating OCPP message action handlers follows the similar approach as in ocpp-library. However, you don\'t need to define classes per OCPP protocol version. \n\n```python\n# provisioning_router.py\nfrom datetime import datetime\n\nfrom ocpp.v16 import call, call_result\nfrom ocpp.v16.enums import Action, RegistrationStatus\n\nfrom ocpp_asgi.router import HandlerContext, Router, Subprotocol\n\nrouter = Router(subprotocol=Subprotocol.ocpp16)\n\n\n@router.on(Action.BootNotification)\nasync def on_boot_notification(\n    *, payload: call.BootNotificationPayload, context: HandlerContext\n) -> call_result.BootNotificationPayload:\n    id = context.charging_station_id\n    print(f"(Central System) on_boot_notification Charging Station {id=}")\n    # Do something with the payload...\n    return call_result.BootNotificationPayload(\n        current_time=datetime.utcnow().isoformat(),\n        interval=10,\n        status=RegistrationStatus.accepted,\n    )\n```\n\n## Central System\n\nCentral System is a collection of routes. You implement it by subclassing from ocpp_asti.ASGIApplication and overriding necessary methods to accommodate your needs. Here\'s a minimal example using uvicorn.\n\n```python\n# central_system.py\nimport uvicorn\nfrom provisioning_router import router\nfrom ocpp_asgi.app import ASGIApplication, RouterContext, Subprotocol\n\n\nclass CentralSystem(ASGIApplication):\n    """Central System is collection of routers."""\n\n\nif __name__ == "__main__":\n    central_system = CentralSystem()\n    central_system.include_router(router)\n    subprotocols = f"{Subprotocol.ocpp16}"\n    headers = [("Sec-WebSocket-Protocol", subprotocols)]\n    uvicorn.run(central_system, host="0.0.0.0", port=9000, headers=headers)\n```\n\n## Run the examples\n\nIn order to run the examples clone the ocpp-asgi repository and install dependencies. Poetry and pyenv are recommended.\n\nThere are two kind of examples how to implement central system with ocpp-asgi: standalone and serverless. Both examples use same ocpp action handlers (routers).\n\n### Standalone example\n\nRun the following commands e.g. in different terminal windows (or run the files within IDE).\n\nStart Central System:\n```\npython ./examples/central_system/standalone/central_system.py\n```\n\nStart Charging Station:\n```\npython ./examples/charging_station.py\n```\n\n### Serverless example\n\nRun the following commands in different terminal windows (or run the files within IDE). Of course when you run the example like this it\'s not really serverless. But deploying something central_system_http.py to e.g. AWS Lambda and running it with Mangum ASGI server is totally possible.\n\nStart Central System HTTP backend:\n```\npython ./examples/central_system/serverless/central_system_http.py\n```\n\nStart Central System WebSocket proxy:\n```\npython ./examples/central_system/serverless/central_system_proxy.py\n```\n\nStart Charging Station:\n```\npython ./examples/charging_station.py\n```',
    'author': 'Ville Kärkkäinen',
    'author_email': 'ville.karkkainen@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/villekr/ocpp-asgi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
